@@ -1,5 +1,5 @@
 /************************************************************************
- * $Id: minibidi.c 9169 2011-05-07 10:57:19Z simon $
+ * $Id: minibidi.c 9426 2012-03-05 18:34:40Z simon $
  *
  * ------------
  * Description:
@@ -14,9 +14,9 @@
  * -----------------
  * Revision Details:    (Updated by Revision Control System)
  * -----------------
- *  $Date: 2011-05-07 11:57:19 +0100 (Sat, 07 May 2011) $
+ *  $Date: 2012-03-05 18:34:40 +0000 (Mon, 05 Mar 2012) $
  *  $Author: simon $
- *  $Revision: 9169 $
+ *  $Revision: 9426 $
  *
  * (www.arabeyes.org - under MIT license)
  *
@@ -58,7 +58,7 @@ shapetypes[(xh)-SHAPE_FIRST].type : SU) /*))*/
 #define leastGreaterEven(x) ( ((x)+2) &~ 1 )
 
 typedef struct bidi_char {
-    wchar_t origwc, wc;
+    unsigned int origwc, wc;
     unsigned short index;
 } bidi_char;
 
@@ -70,7 +70,7 @@ unsigned char setOverrideBits(unsigned char level, unsigned char override);
 int getPreviousLevel(unsigned char* level, int from);
 int do_shape(bidi_char *line, bidi_char *to, int count);
 int do_bidi(bidi_char *line, int count);
-void doMirror(wchar_t* ch);
+void doMirror(unsigned int *ch);
 
 /* character types */
 enum {
@@ -1636,7 +1636,7 @@ int do_bidi(bidi_char *line, int count)
  * takes a pointer to a character that is checked for
  * having a mirror glyph.
  */
-void doMirror(wchar_t* ch)
+void doMirror(unsigned int *ch)
 {
     if ((*ch & 0xFF00) == 0) {
 	switch (*ch) {
