@@ -96,7 +96,7 @@ struct handle_input {
      */
     char buffer[4096];		       /* the data read from the handle */
     DWORD len;			       /* how much data that was */
-    int readret;		       /* lets us know about read errors */
+    int readerr;		       /* lets us know about read errors */
 
     /*
      * Callback function called by this module when data arrives on
@@ -113,7 +113,7 @@ static DWORD WINAPI handle_input_threadfunc(void *param)
     struct handle_input *ctx = (struct handle_input *) param;
     OVERLAPPED ovl, *povl;
     HANDLE oev;
-    int readlen;
+    int readret, readlen;
 
     if (ctx->flags & HANDLE_FLAG_OVERLAPPED) {
 	povl = &ovl;
@@ -132,17 +132,34 @@ static DWORD WINAPI handle_input_threadfunc(void *param)
 	    memset(povl, 0, sizeof(OVERLAPPED));
 	    povl->hEvent = oev;
 	}
-	ctx->readret = ReadFile(ctx->h, ctx->buffer, readlen,
-				&ctx->len, povl);
-	if (povl && !ctx->readret && GetLastError() == ERROR_IO_PENDING) {
+	readret = ReadFile(ctx->h, ctx->buffer,readlen, &ctx->len, povl);
+	if (!readret)
+	    ctx->readerr = GetLastError();
+	else
+	    ctx->readerr = 0;
+	if (povl && !readret && ctx->readerr == ERROR_IO_PENDING) {
 	    WaitForSingleObject(povl->hEvent, INFINITE);
-	    ctx->readret = GetOverlappedResult(ctx->h, povl, &ctx->len, FALSE);
+	    readret = GetOverlappedResult(ctx->h, povl, &ctx->len, FALSE);
+	    if (!readret)
+		ctx->readerr = GetLastError();
+	    else
+		ctx->readerr = 0;
 	}
 
-	if (!ctx->readret)
+	if (!readret) {
+	    /*
+	     * Windows apparently sends ERROR_BROKEN_PIPE when a
+	     * pipe we're reading from is closed normally from the
+	     * writing end. This is ludicrous; if that situation
+	     * isn't a natural EOF, _nothing_ is. So if we get that
+	     * particular error, we pretend it's EOF.
+	     */
+	    if (ctx->readerr == ERROR_BROKEN_PIPE)
+		ctx->readerr = 0;
 	    ctx->len = 0;
+	}
 
-	if (ctx->readret && ctx->len == 0 &&
+	if (readret && ctx->len == 0 &&
 	    (ctx->flags & HANDLE_FLAG_IGNOREEOF))
 	    continue;
 
@@ -227,12 +244,13 @@ struct handle_output {
      * and read by the main thread after receiving that signal.
      */
     DWORD lenwritten;		       /* how much data we actually wrote */
-    int writeret;		       /* return value from WriteFile */
+    int writeerr;		       /* return value from WriteFile */
 
     /*
      * Data only ever read or written by the main thread.
      */
     bufchain queued_data;	       /* data still waiting to be written */
+    enum { EOF_NO, EOF_PENDING, EOF_SENT } outgoingeof;
 
     /*
      * Callback function called when the backlog in the bufchain
@@ -245,11 +263,15 @@ static DWORD WINAPI handle_output_threadfunc(void *param)
 {
     struct handle_output *ctx = (struct handle_output *) param;
     OVERLAPPED ovl, *povl;
+    HANDLE oev;
+    int writeret;
 
-    if (ctx->flags & HANDLE_FLAG_OVERLAPPED)
+    if (ctx->flags & HANDLE_FLAG_OVERLAPPED) {
 	povl = &ovl;
-    else
+	oev = CreateEvent(NULL, TRUE, FALSE, NULL);
+    } else {
 	povl = NULL;
+    }
 
     while (1) {
 	WaitForSingleObject(ctx->ev_from_main, INFINITE);
@@ -257,19 +279,34 @@ static DWORD WINAPI handle_output_threadfunc(void *param)
 	    SetEvent(ctx->ev_to_main);
 	    break;
 	}
-	if (povl)
+	if (povl) {
 	    memset(povl, 0, sizeof(OVERLAPPED));
-	ctx->writeret = WriteFile(ctx->h, ctx->buffer, ctx->len,
-				  &ctx->lenwritten, povl);
-	if (povl && !ctx->writeret && GetLastError() == ERROR_IO_PENDING)
-	    ctx->writeret = GetOverlappedResult(ctx->h, povl,
-						&ctx->lenwritten, TRUE);
+	    povl->hEvent = oev;
+	}
+
+	writeret = WriteFile(ctx->h, ctx->buffer, ctx->len,
+			     &ctx->lenwritten, povl);
+	if (!writeret)
+	    ctx->writeerr = GetLastError();
+	else
+	    ctx->writeerr = 0;
+	if (povl && !writeret && GetLastError() == ERROR_IO_PENDING) {
+	    writeret = GetOverlappedResult(ctx->h, povl,
+					   &ctx->lenwritten, TRUE);
+	    if (!writeret)
+		ctx->writeerr = GetLastError();
+	    else
+		ctx->writeerr = 0;
+	}
 
 	SetEvent(ctx->ev_to_main);
-	if (!ctx->writeret)
+	if (!writeret)
 	    break;
     }
 
+    if (povl)
+	CloseHandle(oev);
+
     return 0;
 }
 
@@ -284,6 +321,11 @@ static void handle_try_output(struct handle_output *ctx)
 	ctx->len = sendlen;
 	SetEvent(ctx->ev_from_main);
 	ctx->busy = TRUE;
+    } else if (!ctx->busy && bufchain_size(&ctx->queued_data) == 0 &&
+               ctx->outgoingeof == EOF_PENDING) {
+        CloseHandle(ctx->h);
+        ctx->h = INVALID_HANDLE_VALUE;
+        ctx->outgoingeof = EOF_SENT;
     }
 }
 
@@ -372,6 +414,7 @@ struct handle *handle_output_new(HANDLE handle, handle_outputfn_t sentdata,
     h->u.o.done = FALSE;
     h->u.o.privdata = privdata;
     bufchain_init(&h->u.o.queued_data);
+    h->u.o.outgoingeof = EOF_NO;
     h->u.o.sentdata = sentdata;
     h->u.o.flags = flags;
 
@@ -380,7 +423,7 @@ struct handle *handle_output_new(HANDLE handle, handle_outputfn_t sentdata,
     add234(handles_by_evtomain, h);
 
     CreateThread(NULL, 0, handle_output_threadfunc,
-		 &h->u.i, 0, &out_threadid);
+		 &h->u.o, 0, &out_threadid);
 
     return h;
 }
@@ -388,11 +431,28 @@ struct handle *handle_output_new(HANDLE handle, handle_outputfn_t sentdata,
 int handle_write(struct handle *h, const void *data, int len)
 {
     assert(h->output);
+    assert(h->u.o.outgoingeof == EOF_NO);
     bufchain_add(&h->u.o.queued_data, data, len);
     handle_try_output(&h->u.o);
     return bufchain_size(&h->u.o.queued_data);
 }
 
+void handle_write_eof(struct handle *h)
+{
+    /*
+     * This function is called when we want to proactively send an
+     * end-of-file notification on the handle. We can only do this by
+     * actually closing the handle - so never call this on a
+     * bidirectional handle if we're still interested in its incoming
+     * direction!
+     */
+    assert(h->output);
+    if (!h->u.o.outgoingeof == EOF_NO) {
+        h->u.o.outgoingeof = EOF_PENDING;
+        handle_try_output(&h->u.o);
+    }
+}
+
 HANDLE *handle_get_events(int *nevents)
 {
     HANDLE *ret;
@@ -512,7 +572,7 @@ void handle_got_event(HANDLE event)
 	    /*
 	     * EOF, or (nearly equivalently) read error.
 	     */
-	    h->u.i.gotdata(h, NULL, (h->u.i.readret ? 0 : -1));
+	    h->u.i.gotdata(h, NULL, -h->u.i.readerr);
 	    h->u.i.defunct = TRUE;
 	} else {
 	    backlog = h->u.i.gotdata(h, h->u.i.buffer, h->u.i.len);
@@ -526,13 +586,13 @@ void handle_got_event(HANDLE event)
 	 * write. Call the callback to indicate that the output
 	 * buffer size has decreased, or to indicate an error.
 	 */
-	if (!h->u.o.writeret) {
+	if (h->u.o.writeerr) {
 	    /*
 	     * Write error. Send a negative value to the callback,
 	     * and mark the thread as defunct (because the output
 	     * thread is terminating by now).
 	     */
-	    h->u.o.sentdata(h, -1);
+	    h->u.o.sentdata(h, -h->u.o.writeerr);
 	    h->u.o.defunct = TRUE;
 	} else {
 	    bufchain_consume(&h->u.o.queued_data, h->u.o.lenwritten);
